# Drag and drop for vanilla Tk4.0

# See the interface and the demo for details on how to use

# Set the next var to turn on debugging. NB. There's _lots_ of it...
if [catch {set dragdrop(debug)}] {
    set dragdrop(debug)      0
}
# I'm allowed to drop here
if [catch {set dragdrop(yes)}] {
    set dragdrop(yes)    [list arrow green1 green4]
}
# I'm not allowed to drop here.
if [catch {set dragdrop(no)}] {
    set dragdrop(no)     [list circle red1 red4]
}

# DO NOT EDIT UNLESS YOU KNOW WHAT YOU ARE DOING! Or you feel like
# tracking down bugs... :^)
set dragdrop(types)      {}
set dragdrop(source)     {}
set dragdrop(widget)     {}
set dragdrop(cursor)     {}
set dragdrop(rmt_widget) {}
set dragdrop(rmt_interp) {}

### ----------------------------------------------------------------------
#
# Transmitting end
#

# Finds the name and widget_id of a drag-and-drop supporting Tk4 app
# at the current mouse location, or returns an empty string.
#
# I can blithely use catch in this proc, since winfo containing
# shouldn't raise any errors except for ones that would have indicated
# that the drag-and-drop wouldn't have worked anyway. info globals
# should never fail, especially if winfo containing works... :^)
;proc dragdrop_find {x y} {
    global dragdrop

    # First check to see if the cached interpreter name is still the
    # right one. This should cut down on needless searching...
    if [llength $dragdrop(rmt_interp)] {
	if ![catch {
	    # Is there a Tk app at that point?
	    set wid [send -- $dragdrop(rmt_interp) winfo containing $x $y]
	}] {
	    return [list $dragdrop(rmt_interp) $wid]
	}
    }	

    # OK. _Now_ we search...
    foreach i [winfo interps] {
	if ![catch {
	    # Is there a Tk app at that point?
	    set wid [send -- $i winfo containing $x $y]
	}] {
	    # Does it support drag-and-drop?
	    if [llength [send -- $i info globals dragdrop]] {
		return [list $i $wid]
	    }
	}
    }

    # Couldn't find anything...
    return {}
}

# Sets up what data you are going to drag about.
# Attach to <1> (or <2> or <3>)
proc drag_start {x y types} {
    global dragdrop dragdropnotify

    set dragdrop(widget) [winfo containing $x $y]
    if $dragdrop(debug) {
	puts "[tk appname]$dragdrop(widget) started drag at ($x,$y)"
    }
    set dragdrop(types) $types
    set dragdrop(cursor) [$dragdrop(widget) cget -cursor]
    catch {
	if [llength $dragdropnotify($dragdrop(widget))] {
	    $dragdropnotify($dragdrop(widget)) startat $dragdrop(widget) $x $y
	}
    } m
    if {$dragdrop(debug) && [string length $m]} {
	puts "notify returned >>$m<<"
    }
    # Update the cursor by dragging to here... (Efficient? No. Easy? Yes! :^)
    drag $x $y
}

# Informs the far end that the cursor is where it is.
# Attach to <B1-Motion> in _transmitter_
proc drag {x y} {
    global dragdrop dragdropnotify

    set iwid [dragdrop_find $x $y]
    set mywidget $dragdrop(widget)
    if [llength $iwid] {
	set i [lindex $iwid 0]
	set wid [lindex $iwid 1]

	set ready4drop 0
	if {[string compare $wid $dragdrop(rmt_widget)] || \
		[string compare $i $dragdrop(rmt_interp)]} {
	    catch {
		send -- $i [list drag_at {} {} {}]
	    }
	}
	catch {
	    set ready4drop [send -- $i [list drag_at $x $y [tk appname]]]
	}
	if $ready4drop {
	    $mywidget configure -cursor $dragdrop(yes)
	    set dragdrop(rmt_widget) $wid
	    set dragdrop(rmt_interp) $i
	    catch {
		if [llength $dragdropnotify($mywidget)] {
		    $dragdropnotify($mywidget) able $mywidget $x $y $i $wid
		}
	    } m
	} else {
	    $mywidget configure -cursor $dragdrop(no)
	    set dragdrop(rmt_widget) {}
	    set dragdrop(rmt_interp) {}
	    catch {
		if [llength $dragdropnotify($mywidget)] {
		    $dragdropnotify($mywidget) unable $mywidget $x $y
		}
	    } m
	}
	if {$dragdrop(debug) && [string length $m]} {
	    puts "notify returned >>$m<<"
	}
	return
    }
    $mywidget configure -cursor $dragdrop(no)
    if [llength $dragdrop(rmt_interp)] {
	catch {
	    send -- $i [list drag_at {} {} {}]
	}
    }
    set dragdrop(rmt_widget) {}
    set dragdrop(rmt_interp) {}
    catch {
	if [llength $dragdropnotify($mywidget)] {
	    $dragdropnotify($mywidget) unable $mywidget $x $y
	}
    } m
    if {$dragdrop(debug) && [string length $m]} {
	puts "notify returned >>$m<<"
    }
}

# Drops the data (if possible).
# Attach to <ButtonRelease-1> in _transmitter_
proc drop {x y} {
    global dragdrop dragdropnotify

    set flag 0
    set iwid [dragdrop_find $x $y]
    if [llength $iwid] {
	set i [lindex $iwid 0]
	set wid [lindex $iwid 1]

	set canIgo [send -- $i [list drag_at $x $y [tk appname]]]
	if $canIgo {
	    catch {
		send -- $i [list drag_at {} {} {}]
		send -- $i [list drop_at $x $y [tk appname]]
		if $dragdrop(debug) {
		    puts "drop succeeded (sent to $i)"
		}
		$dragdrop(widget) configure -cursor $dragdrop(cursor)
	    }
	    catch {
		if [llength $dragdropnotify($dragdrop(widget))] {
		    $dragdropnotify($dragdrop(widget)) dropped $dragdrop(widget) $x $y $i $wid
		}
	    } m
	    if {$dragdrop(debug) && [string length $m]} {
		puts "notify returned >>$m<<"
	    }
	    return
	} else {
	    send -- $i [list drag_at {} {} {}]
	}
    }
    $dragdrop(widget) configure -cursor $dragdrop(cursor)
    if $dragdrop(debug) {
	puts "drop failed. Not over a Tk app?"
    }
    set dragdrop(types) {}
    catch {
	if [llength $dragdropnotify($dragdrop(widget))] {
	    $dragdropnotify($dragdrop(widget)) failed $dragdrop(widget) $x $y
	}
    } m
    if {$dragdrop(debug) && [string length $m]} {
	puts "notify returned >>$m<<"
    }
}

proc drop_get_data {type} {
    global dragdrop dragserver
    set retval {}
    catch {
	set retval [$dragserver($dragdrop(widget)) $dragdrop(widget) $type]
    }
    if $dragdrop(debug) {
	puts "Serving data >>$retval<< for $type request to [tk appname]$dragdrop(widget)"
    }
    return $retval
}

### ----------------------------------------------------------------------
#
# Receiving end
#

# Handles dragging signals.
proc drag_at {x y dragdatasource} {
    global draghandler dragdrop
    if [catch {
	set wid [winfo containing $x $y]
    }] {
	if $dragdrop(debug) {
	    puts "[tk appname]$dragdrop(rmt_widget) clear drag"
	}
	catch {
	    $draghandler($dragdrop(rmt_widget)) dragdrop(rmt_widget) \
		    {} {} $dragdrop(source) $dragdrop(types)
	}
	set dragdrop(source) {}
    } else {
	set dragdrop(rmt_widget) $wid
	if $dragdrop(debug) {
	    puts "[tk appname]$wid drag at ($x,$y)"
	}
	if [catch {
	    set draghandl $draghandler($wid)
	}] {
	    # Not a data destination
	    return 0
	} else {
	    if [string compare $dragdatasource $dragdrop(source)] {
		set dragdrop(types) [send -- $dragdatasource set dragdrop(types)]
		set dragdrop(source) $dragdatasource
	    }
	    set retval 0
	    if [catch {
		set retval [$draghandl $wid [expr $x-[winfo rootx $wid]] \
			[expr $y-[winfo rooty $wid]] $dragdrop(source) $dragdrop(types)]
	    } m] {
		if $dragdrop(debug) {
		    puts $m
		}
	    }
	    # Is a data destination? Truth for yes, falsehood for no.
	    return $retval
	}
    }
}

# Handles dropped data. _Runs_in_receiver_ and dispatches the handler
proc drop_at {x y dragdatasource} {
    global drophandler dragdrop dropacceptedtypes
    set wid [winfo containing $x $y]
    if $dragdrop(debug) {
	puts "[tk appname]$wid received drop at ($x,$y)"
    }
    if [catch {
	set drophandl $drophandler($wid)
    }] {
	if $dragdrop(debug) {
	    puts "No handler for $wid"
	}
    } else {
	if [string compare $dragdatasource $dragdrop(source)] {
	    set dragdrop(types) [send -- $dragdatasource set dragdrop(types)]
	    set dragdrop(source) $dragdatasource
	}
	foreach dat $dropacceptedtypes($wid) {
	    if $dragdrop(debug) {
		puts "Trying $dat (In $dragdrop(types)?)"
	    }
	    if {[lsearch -exact $dragdrop(types) $dat] >= 0} {
		set data [send -- $dragdatasource drop_get_data $dat]
		if [catch {
		    $drophandl $wid [expr $x-[winfo rootx $wid]] \
			    [expr $y-[winfo rooty $wid]] $dragdrop(source) $dat $data
		} m] {
		    if $dragdrop(debug) {
			puts >>$m<<
		    }
		}
		break
	    } else {
		if $dragdrop(debug) {
		    puts "No handler for $dat"
		}
	    }
	}
	set dragdrop(source) {}
	set dragdrop(types) {}
    }
}
